/*
 * FILE:
 *   network_connection.cpp
 *
 * AUTHOR:
 *   Stephen Thompson, 2008
 *
 * COPYRIGHT:
 *   Public Domain
 *
 */

#include "network_connection.hpp"
#include "../core/coercri_error.hpp"

namespace Coercri {
    
    unsigned char NetworkConnection::getUbyte()
    {
        if (incoming_index < incoming_packet.size()) {
            const unsigned char result = incoming_packet[incoming_index++];
            if (incoming_index == incoming_packet.size()) {
                incoming_packet.clear();
                incoming_index = 0;
            }
            return result;
        } else {
            throw CoercriError("Unexpected end of network packet");
        }
    }

    signed char NetworkConnection::getByte()
    {
        int x = getUbyte();
        if (x < 128) {
            return x;
        } else {
            return x - 256;
        }
    }

    unsigned short int NetworkConnection::getUshort()
    {
        int high = getUbyte();
        int low = getUbyte();
        return (high * 256) + low;
    }

    short int NetworkConnection::getShort()
    {
        int high = getByte();
        int low = getUbyte();
        return (high * 256 + low);
    }

    unsigned int NetworkConnection::getUlong()
    {
        int b3 = getUbyte();
        int b2 = getUbyte();
        int b1 = getUbyte();
        int b0 = getUbyte();
        return (b3*16777216 + b2*65536 + b1*256 + b0);
    }

    int NetworkConnection::getLong()
    {
        int b3 = getByte();
        int b2 = getUbyte();
        int b1 = getUbyte();
        int b0 = getUbyte();
        return (b3*16777216 + b2*65536 + b1*256 + b0);
    }

    int NetworkConnection::getCInt()
    {
        unsigned int val;
        bool sign;
        unsigned char c;
        c = getUbyte();
        // bit 7 = extend bit
        // bit 6 = sign bit
        // bits 5-0 = least significant bits
        val = (c & 0x3f);
        sign = (c & 0x40);
        if (c & 0x80) {
            // 2nd byte
            c = getUbyte();
            val |= ((c & 0x7f) << 6);
            if (c & 0x80) {
                // 3rd byte
                c = getUbyte();
                val |= ((c & 0x7f) << 13);
                if (c & 0x80) {
                    // 4th byte
                    c = getUbyte();
                    val |= (c << 20);
                } 
            }
        }
        if (sign) {
            return -val-1;
        } else {
            return val;
        }
    }

    std::string NetworkConnection::getString()
    {
        std::string s;
        int size = getCInt();
        s.resize(size);
        for (std::string::iterator it = s.begin(); it != s.end(); ++it) {
            *it = getByte();
        }
        return s;
    }

    void NetworkConnection::putUbyte(unsigned char c)
    {
        outgoing_packet.push_back(c);
    }

    void NetworkConnection::putByte(signed char c)
    {
        outgoing_packet.push_back(c);
    }

    void NetworkConnection::putUshort(unsigned short int s)
    {
        outgoing_packet.push_back(s >> 8);
        outgoing_packet.push_back(s & 0xff);
    }

    void NetworkConnection::putShort(short int s)
    {
        // converting signed to unsigned is safe
        putUshort(s);
    }

    void NetworkConnection::putUlong(unsigned int x)
    {
        outgoing_packet.push_back(x >> 24);
        outgoing_packet.push_back((x >> 16) & 0xff);
        outgoing_packet.push_back((x >> 8) & 0xff);
        outgoing_packet.push_back(x & 0xff);
    }

    void NetworkConnection::putLong(int x)
    {
        putUlong(x);
    }

    void NetworkConnection::putCInt(int x)
    {
        if (x < -268435456 || x > 268435455) {
            throw CoercriError("NetworkConnection::putCInt: out of range");
        }
        
        unsigned int sign;
        unsigned int val;
        if (x >= 0) {
            val = x;
            sign = 0;
        } else {
            val = -x-1;
            sign = 0x40;
        }

        if (val <= 63) {
            putUbyte(sign + val);
        } else {
            putUbyte(0x80 + sign + (val & 0x3f));
            if (val <= 8191) {
                putUbyte(val >> 6);
            } else {
                putUbyte(0x80 + ((val >> 6) & 0x7f));
                if (val <= 1048575) {
                    putUbyte(val >> 13);
                } else {
                    putUbyte(0x80 + ((val >> 13) & 0x7f));
                    putUbyte(val >> 20);
                }
            }
        }
    }

    void NetworkConnection::putString(const std::string &s)
    {
        putCInt(s.size());
        for (std::string::const_iterator it = s.begin(); it != s.end(); ++it) {
            putUbyte(*it);
        }
    }

    void NetworkConnection::endPacket()
    {
        if (!outgoing_packet.empty()) {
            transmit(outgoing_packet);
        }
        outgoing_packet.clear();
    }

    bool NetworkConnection::inAvail()
    {
        if (incoming_packet.empty()) {
            receive(incoming_packet);
        }
        return !incoming_packet.empty();
    }
}

