/*
 * connecting_screen.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "connecting_screen.hpp"
#include "error_screen.hpp"
#include "knights_game.hpp"
#include "loading_screen.hpp"
#include "net_msgs.hpp"
#include "title_screen.hpp"
#include "x_centre.hpp"

#include "gui_button.hpp"
#include "gui_centre.hpp"
#include "gui_panel.hpp"

#include "coercri/core/coercri_error.hpp"
#include "coercri/gfx/font.hpp"
#include "coercri/gfx/gfx_context.hpp"
#include "coercri/gfx/color.hpp"
#include "coercri/gfx/window_listener.hpp"
#include "coercri/network/basic_socket.hpp"

#include "boost/scoped_ptr.hpp"

#include <set>
using namespace std;

class ConnectingScreenImpl : public gcn::ActionListener {
public:
    ConnectingScreenImpl(KnightsGame &kg, gcn::Gui &gui, const string &addr);
    void action(const gcn::ActionEvent &event);

    KnightsGame & getKnightsGame() const { return knights_game; }
    Coercri::BasicSocket & getSocket() const { return *socket; }
    
private:
    KnightsGame &knights_game;
    string address;
    
    scoped_ptr<GuiCentre> centre;
    scoped_ptr<GuiPanel> panel;
    scoped_ptr<gcn::Container> container;
    scoped_ptr<gcn::Label> label;
    scoped_ptr<gcn::Button> cancel_button;

    shared_ptr<Coercri::BasicSocket> socket;
};

ConnectingScreenImpl::ConnectingScreenImpl(KnightsGame &kg, gcn::Gui &gui, const string &addr)
    : knights_game(kg), address(addr)
{
    container.reset(new gcn::Container);
    container->setOpaque(false);
    
    const int pad = 10;
    int y = pad;
    
    if (knights_game.server) {
        label.reset(new gcn::Label("Waiting for connection"));
    } else {
        label.reset(new gcn::Label("Connecting to " + address + "..."));
    }
    container->add(label.get(), pad, y);
    y += label->getHeight() + pad;
    
    cancel_button.reset(new GuiButton("Cancel"));
    cancel_button->addActionListener(this);
    container->add(cancel_button.get(), pad + label->getWidth() / 2 - cancel_button->getWidth() / 2, y);

    container->setSize(2*pad + label->getWidth(), y + cancel_button->getHeight() + pad);
    
    panel.reset(new GuiPanel(container.get()));
    centre.reset(new GuiCentre(panel.get()));
    gui.setTop(centre.get());

    socket = kg.net_driver->createBasicSocket(CfgInt("broadcast_port"));
}

void ConnectingScreenImpl::action(const gcn::ActionEvent &event)
{
    if (event.getSource() == cancel_button.get()) {
        // Go back to title screen
        auto_ptr<Screen> title_screen(new TitleScreen);
        knights_game.requestScreenChange(title_screen);
    }
}

ConnectingScreen::ConnectingScreen()
{
}

ConnectingScreen::ConnectingScreen(const string &a)
    : address(a)
{
}

bool ConnectingScreen::start(KnightsGame &kg, shared_ptr<Coercri::Window> win, gcn::Gui &gui)
{
    pimpl.reset(new ConnectingScreenImpl(kg, gui, address));
    
    if (kg.server) {
        // Enable Server
        try {
            kg.net_driver->enableServer(true);
        } catch (const Coercri::CoercriError &) {
            // Coercri could not enable the network server for us. Most likely the port is already in use.
            auto_ptr<Screen> error_screen(new ErrorScreen("Failed to create server"));
            kg.requestScreenChange(error_screen);
        }
    } else {
        // Create the connection to the server
        kg.remote = kg.net_driver->openConnection(address, CfgInt("port_number"));
    }
    
    return true;
}

void ConnectingScreen::update()
{
    bool done = false;

    if (pimpl->getKnightsGame().server) {

        if (!pimpl->getKnightsGame().remote) {
            // Nobody has connected yet
            // Listen for incoming connections
            Coercri::NetworkDriver::Connections new_connections
                = pimpl->getKnightsGame().net_driver->pollIncomingConnections();
            if (!new_connections.empty()) {
                // Somebody has connected
                ASSERT(new_connections.size() == 1);
                pimpl->getKnightsGame().remote = new_connections.front();
            }
        }

        if (pimpl->getKnightsGame().remote && pimpl->getKnightsGame().remote->inAvail()) {
            // Someone has connected. Check their version string
            const string other_ver_string = pimpl->getKnightsGame().remote->getString();
            if (other_ver_string != string(CONNECTION_MSG)) {
                // refuse the connection
                pimpl->getKnightsGame().remote->close();
                // go to error screen
                auto_ptr<Screen> error_screen(new ErrorScreen("Protocol Error"));
                pimpl->getKnightsGame().requestScreenChange(error_screen);
            } else {
                // We are done
                done = true;
            }
        }

        // Check for incoming ping messages on the broadcast-socket. Send replies if necessary.
        string msg, address;
        // Don't send more than one reply to the same address...
        set<string> known_addresses;
        while (pimpl->getSocket().receive(address, msg)) {
            if (msg == BROADCAST_PING_MSG && known_addresses.find(address) == known_addresses.end()) {
                pimpl->getSocket().send(address, BROADCAST_PONG_MSG);
                known_addresses.insert(address);
            }
        }
        
    } else {
        Coercri::NetworkConnection::State state = pimpl->getKnightsGame().remote->getState();
        if (state == Coercri::NetworkConnection::CLOSED) {
            // Our connection timed out, or was refused. Go to error screen
            auto_ptr<Screen> error_screen(new ErrorScreen("Connection Failed"));
            pimpl->getKnightsGame().requestScreenChange(error_screen);
        } else if (state == Coercri::NetworkConnection::CONNECTED) {
            // Send our version string.
            pimpl->getKnightsGame().remote->putString(CONNECTION_MSG);
            pimpl->getKnightsGame().remote->endPacket();
            // Let the game begin!
            done = true;
        }
        // Otherwise -- we are in PENDING state. Keep waiting for the connection.
    }

    if (done) {
        // Go to the loading screen next.
        auto_ptr<Screen> loading_screen(new LoadingScreen);
        pimpl->getKnightsGame().requestScreenChange(loading_screen);
    }
}


