/*
 * game_client.hpp
 *
 * GameClient is responsible for receiving network packets and changing them
 * into game commands.
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#ifndef GAME_CLIENT_HPP
#define GAME_CLIENT_HPP

#include "map_support.hpp"

#include "coercri/gfx/graphic.hpp"
#include "coercri/network/network_connection.hpp"

#include <utility>
using namespace std;

class Anim;
class ClientConfig;
class Control;
class LocalDungeonView;
class Overlay;


class GameClient {
public:
    // Note: Ctor assumes that we have received all config data already, and that
    // the stream is clear for sending/receiving game data. (Nothing is sent or received
    // until the first call to update, though.)
    GameClient(Coercri::NetworkConnection &connection_to_server, LocalDungeonView &client_dview,
               ClientConfig &conf);

    // Update by moving GVT forward, and then receiving commands from the network.
    // Note we accept a *delta* to be added to GVT, not an absolute new GVT.
    void update(int delta_gvt);

    // Query whether the game is over (we have received a C_WIN_GAME, C_LOSE_GAME or
    // C_QUIT_GAME).
    bool isGameOver() const { return game_over; }
    bool isGameQuit() const { return game_quit; }

    // Send a message to the server, asking it to quit the game (ie asking it to send us a C_QUIT_GAME)
    // Used when the client presses ESCAPE in-game.
    void sendQuitRequest();
    
    
    // get gvt
    int getGVT() const { return gvt; }
    
private:
    int recvCommandNo();
    int recvRoomNo() { return remote.getUbyte(); }
    MapCoord recvMapCoord();
    void recvRoomCoord(int &x, int &y) { recvNibbles(x, y); x--; y--; }
    int recvSpeed() { return remote.getUshort(); }
    void recvTileInfo(int &d, int &c, bool &cc);
    const Anim * recvAnimPtr();
    const Control * recvControlPtr();
    pair<const Control*, bool> recvControlPlusBool();
    const Coercri::Graphic * recvGraphicPtr();
    const Overlay * recvOverlayPtr();
    string recvString();
    void recvNibbles(int &, int &);
    void sendGameOverMsg();  // this also does an endPacket()
    void sendControl(const Control *ctrl);
    void receiveCommands();

private:
    Coercri::NetworkConnection &remote;
    LocalDungeonView &dview;
    ClientConfig &conf;
    int gvt;
    const Control *last_sent_ctrl;
    bool game_over, game_quit;
};

#endif
