/*
 * gfx_manager.hpp
 *
 * Provides some additional graphics functionality including resizing,
 * recolouring, and serialization.
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#ifndef GFX_MANAGER_HPP
#define GFX_MANAGER_HPP

#include "colour_change.hpp"

#include "coercri/gfx/font.hpp"
#include "coercri/gfx/gfx_context.hpp"
#include "coercri/gfx/gfx_driver.hpp"
#include "coercri/gfx/graphic.hpp"
#include "coercri/gfx/ttf_loader.hpp"
#include "coercri/network/network_connection.hpp"

#include "boost/shared_ptr.hpp"
using namespace boost;

#include <list>
#include <map>
using namespace std;

class ColourChange;
class GfxResizer;

class GfxManager {
public:
    GfxManager(shared_ptr<Coercri::GfxDriver> gfx_driver_,
               shared_ptr<Coercri::TTFLoader> ttf_loader_,
               const vector<string> &ttf_font_names_,
               const vector<string> &bmp_font_names_);
    ~GfxManager();
    
    // Fonts and text

    shared_ptr<Coercri::Font> getFont();
    void setFontSize(int new_size);
    
    
    // Graphics loading and serialization.
    // 
    // Notes:
    // 
    // 1) serializeGraphic can only be called on graphics that were
    // created by GfxManager::createGraphicFromFile.
    // 
    // 2) getGraphicID only works on serialized graphics.
    // 
    // 3) Graphics returned by the create methods shouldn't be deleted
    // by the caller -- they will be freed automatically by
    // ~GfxManager(), or can be freed manually by calling
    // deleteAllGraphics().
    
    void serializeGraphic(Coercri::NetworkConnection &remote,
                          const Coercri::Graphic &g, int id);  // sets the graphic id at the same time
    Coercri::Graphic * createGraphicFromNetwork(Coercri::NetworkConnection &remote);
    int getGraphicID(const Coercri::Graphic &g) const;
    Coercri::Graphic * createGraphicFromFile(const string &filename, int hx, int hy, int r, int g, int b);
    void deleteAllGraphics();

    // Drawing resized and/or colour-changed graphics.
    
    void setGfxResizer(shared_ptr<GfxResizer> resizer);
    shared_ptr<GfxResizer> getGfxResizer();
    void drawTransformedGraphic(Coercri::GfxContext &gc, int x, int y,
                                const Coercri::Graphic &gfx, int new_width, int new_height);
    void drawTransformedGraphic(Coercri::GfxContext &gc, int x, int y,
                                const Coercri::Graphic &gfx, int new_width, int new_height,
                                const ColourChange &cc);

private:
    // typedefs, structs
    struct GraphicSpec {
        string filename;
        int hx, hy;
        int r, g, b;
        bool operator<(const GraphicSpec &rhs) const {
            return
                filename < rhs.filename ? true :
                filename > rhs.filename ? false :
                hx < rhs.hx ? true :
                hx > rhs.hx ? false :
                hy < rhs.hy ? true :
                hy > rhs.hy ? false :
                r < rhs.r ? true :
                r > rhs.r ? false :
                g < rhs.g ? true :
                g > rhs.g ? false :
                b < rhs.b ? true :
                false;
        }
    };
    
    typedef map<GraphicSpec, shared_ptr<Coercri::Graphic> > GfxMap;
    typedef map<const Coercri::Graphic*, GraphicSpec> GfxRevMap;
    typedef map<const Coercri::Graphic*, int> IDMap;

    struct GraphicKey {
        const Coercri::Graphic * original;
        ColourChange cc;
        int new_width, new_height;

        bool operator<(const GraphicKey &rhs) const
        {
            return
                original < rhs.original ? true :
                rhs.original < original ? false :
                cc < rhs.cc ? true :
                rhs.cc < cc ? false :
                new_width < rhs.new_width ? true :
                rhs.new_width < new_width ? false :
                new_height < rhs.new_height;
        }
    };

    typedef list<GraphicKey> RecentList;  // MRU at front, LRU at back

    struct GraphicData {
        shared_ptr<Coercri::Graphic> new_graphic;
        RecentList::iterator list_position;
    };

    typedef map<GraphicKey, GraphicData> CachedGfxMap;
        
private:
    shared_ptr<Coercri::Graphic> createGraphic(const GraphicKey &key);
    const Coercri::Graphic & getGraphic(const GraphicKey &key);
    bool tooBig() const;
    void deleteOld();
    const Coercri::Graphic & getGraphicWithCC(const Coercri::Graphic &original, const ColourChange &cc);
    const Coercri::Graphic & getResizedGraphic(const Coercri::Graphic &original, int new_width, int new_height);
    
private:
    shared_ptr<Coercri::GfxDriver> gfx_driver;
    shared_ptr<GfxResizer> gfx_resizer;

    // Font data
    shared_ptr<Coercri::TTFLoader> ttf_loader;
    vector<string> ttf_font_names, bmp_font_names;
    int font_size;
    shared_ptr<Coercri::Font> font;

    // Filename <-> graphic map.
    GfxMap gfx_map;
    GfxRevMap gfx_rev_map;

    // Graphic -> ID map
    IDMap id_map;
    
    // Caches for transformed graphics
    CachedGfxMap cached_gfx_map;
    RecentList recent_list;
    int npix;
};

#endif
