/*
 * knights_game.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "control_actions.hpp"
#include "error_screen.hpp"
#include "gfx_manager.hpp"
#include "gfx_resizer_compose.hpp"
#include "gfx_resizer_nearest_nbr.hpp"
#include "gfx_resizer_scale2x.hpp"
#include "keyboard_controller.hpp"
#include "knights_game.hpp"
#include "load_font.hpp"
#include "menu_selections.hpp"
#include "misc/my_exceptions.hpp"
#include "options.hpp"
#include "rstream/rstream.hpp"
#include "sound_manager.hpp"
#include "title_screen.hpp"

#include "kconfig/kconfig_loader.hpp"

#include "coercri/enet/enet_network_driver.hpp"
#include "coercri/gcn/cg_font.hpp"
#include "coercri/gcn/cg_listener.hpp"
#include "coercri/gfx/window_listener.hpp"
#include "coercri/sdl/gfx/sdl_gfx_driver.hpp"
#include "coercri/sdl/gfx/sdl_ttf_loader.hpp"
#include "coercri/sdl/sound/sdl_sound_driver.hpp"
#include "coercri/sdl/timer/sdl_timer.hpp"

#include "guichan.hpp"

#include "boost/scoped_ptr.hpp"
using namespace boost;

#include <cstdlib>

#ifdef WIN32
#include <shlobj.h>
#endif


/////////////////////////////////////////////////////////////
// WindowListener to check for window close events
/////////////////////////////////////////////////////////////

namespace {
    class WindowCloseListener : public Coercri::WindowListener {
    public:
        explicit WindowCloseListener(KnightsGame &k) : kg(k) { }
        void onClose() { kg.requestQuit(); }
    private:
        KnightsGame &kg;
    };
}


/////////////////////////////////////////////////////////////
// Definition of KnightsGameImpl
/////////////////////////////////////////////////////////////

struct KnightsGameImpl {
    shared_ptr<Coercri::GfxDriver> gfx_driver;
    shared_ptr<Coercri::SoundDriver> sound_driver;
    shared_ptr<Coercri::Timer> timer;
    shared_ptr<Coercri::TTFLoader> ttf_loader;
    shared_ptr<Coercri::Window> window;
    shared_ptr<WindowCloseListener> wcl;

    shared_ptr<gcn::Gui> gui;
    shared_ptr<Coercri::CGListener> cg_listener;

    // we do not want these shared, we want to be sure there is only one copy of each:
    auto_ptr<Screen> current_screen, requested_screen;
    
    unsigned int last_time;
    unsigned int update_interval;
    
    bool running;

    shared_ptr<GfxManager> gfx_manager;
    shared_ptr<SoundManager> sound_manager;
    shared_ptr<Controller> left_controller, right_controller;

    shared_ptr<Coercri::Font> font;
    scoped_ptr<gcn::Font> gcn_font;

    scoped_ptr<Options> options;
    std::string options_filename;
    
    KnightsGameImpl() : last_time(0), update_interval(0), running(true) { }
};


/////////////////////////////////////////////////////
// Constructor (One-Time Initialization)
/////////////////////////////////////////////////////

KnightsGame::KnightsGame(const string &resource_dir)
    : pimpl(new KnightsGameImpl)
{
    const char * game_name = "Knights";

    // initialize resource lib
    RStream::Initialize(resource_dir);

    // initialize "Cfg" library
    CfgInit();

    // initialize game options
    pimpl->options.reset(new Options);
#ifdef WIN32
    TCHAR szPath[MAX_PATH];
    if(SUCCEEDED(SHGetFolderPath(NULL, 
                                 CSIDL_APPDATA,
                                 NULL,
                                 0,
                                 szPath))) {
        pimpl->options_filename = szPath;
        pimpl->options_filename += "/knights_config.txt";
    }
#else
    // assume getenv("HOME") will work.
    pimpl->options_filename = std::getenv("HOME");
    pimpl->options_filename += "/.knights_config";
#endif
    if (!pimpl->options_filename.empty()) {
        ifstream str(pimpl->options_filename.c_str());
        *pimpl->options = LoadOptions(str);
    }
    
    
    // Set up Coercri
    // (Here we use SDL, but a different backend could be selected
    // just by changing the following few lines.)
    pimpl->gfx_driver.reset(new Coercri::SDLGfxDriver);

#ifndef DISABLE_SOUND
    try {
        pimpl->sound_driver.reset(new Coercri::SDLSoundDriver(CfgFloat("sound_volume") / 100.0f));
    } catch (std::exception &e) {
        // Print warning message to cout, and continue without a sound_driver
        std::cout << "Problem initializing audio: " << e.what() << std::endl;
    }
#endif
    
    pimpl->timer.reset(new Coercri::SDLTimer);
    pimpl->ttf_loader.reset(new Coercri::SDLTTFLoader);

    // use the enet network driver
    net_driver.reset(new Coercri::EnetNetworkDriver(1, 1));
    net_driver->enableServer(false);  // start off disabled.
    net_driver->setServerPort(CfgInt("port_number"));

    // Set the Windows resource number for the window icon
    pimpl->gfx_driver->setWindowsIcon(1);
    
    // Open the game window.
    bool fullscreen = pimpl->options->fullscreen;
    int width, height;
    if (fullscreen) {
        getDesktopResolution(width, height);
    } else {
        width = DefaultWindowWidth();
        height = DefaultWindowHeight();
    }
    pimpl->window = pimpl->gfx_driver->createWindow(width, height, true, fullscreen, game_name);
        
    // Set up the standard controls
    SetupStandardControls();

    // Get the font names.
    // NOTE: we always try to load TTF before bitmap fonts, irrespective of the order they are in the file.
    vector<string> ttf_font_names, bitmap_font_names;
    {
        RStream str("fonts.txt");

        while (str) {   
            string x;
            getline(str, x);
        
            // Left trim
            x.erase(x.begin(), find_if(x.begin(), x.end(), not1(ptr_fun<int,int>(isspace))));
            // Right trim
            x.erase(find_if(x.rbegin(), x.rend(), not1(ptr_fun<int,int>(isspace))).base(), x.end());
            
            if (x.empty()) continue;
        
            if (x[0] == '+') bitmap_font_names.push_back(x.substr(1));
            else if (x[0] != '#') ttf_font_names.push_back(x);
        }
    }
    
    // Font for the gui
    pimpl->font = LoadFont(*pimpl->ttf_loader, ttf_font_names, bitmap_font_names, 15);

    // Setup gfx & sound managers
    pimpl->gfx_manager.reset(new GfxManager(pimpl->gfx_driver, pimpl->ttf_loader, ttf_font_names, bitmap_font_names));
    setupGfxResizer();
    pimpl->sound_manager.reset(new SoundManager(pimpl->sound_driver));

    // Setup Controllers
    setupControllers();

    // Set up Guichan and CGListener
    // NOTE: No need to set Input and Graphics for guichan since this is handled by CGListener.
    pimpl->gui.reset(new gcn::Gui);
    pimpl->cg_listener.reset(new Coercri::CGListener(pimpl->window, pimpl->gui, pimpl->timer));
    pimpl->window->addWindowListener(pimpl->cg_listener.get());

    // Add the WindowCloseListener
    pimpl->wcl.reset(new WindowCloseListener(*this));
    pimpl->window->addWindowListener(pimpl->wcl.get());
    
    // Set guichan's global widget font
    pimpl->gcn_font.reset(new Coercri::CGFont(pimpl->font, Coercri::Color(0,0,0)));
    gcn::Widget::setGlobalFont(pimpl->gcn_font.get());
}


//////////////////////////////////////////////
// Screen handling
//////////////////////////////////////////////

void KnightsGame::requestScreenChange(auto_ptr<Screen> screen)
{
    pimpl->requested_screen = screen;
}

void KnightsGame::executeScreenChange()
{
    if (pimpl->requested_screen.get() != pimpl->current_screen.get() && pimpl->requested_screen.get()) {
        // Close the gui
        pimpl->gui->setTop(0);
        pimpl->cg_listener->disableGui();

        // Swap screens
        pimpl->current_screen = pimpl->requested_screen;   // clears requested_screen

        // Initialize the new screen, and bring up the gui if required
        const bool requires_gui = pimpl->current_screen->start(*this, pimpl->window, *pimpl->gui);
        if (requires_gui) pimpl->cg_listener->enableGui();

        // Work around bug in guichan where Button::mHasMouse will be set to false for new buttons
        // even if the mouse is actually within the button's area when it is created.
        repeatLastMouseInput();
        
        // Reset timer.
        pimpl->last_time = pimpl->timer->getMsec();
        pimpl->update_interval = pimpl->current_screen->getUpdateInterval();

        // Ensure the screen gets repainted
        pimpl->window->invalidateAll();
    }
}

void KnightsGame::requestQuit()
{
    pimpl->running = false;
}

void KnightsGame::repeatLastMouseInput()
{
    pimpl->cg_listener->repeatLastMouseInput();
}

//////////////////////////////////////////////
// Game state handling
//////////////////////////////////////////////

void KnightsGame::resetAll()
{
    // Clean up after an exit back to title screen.
    if (remote) remote->close();
    remote.reset();
    net_driver->enableServer(false);
    server_config.reset();
    client_config.reset();
    pimpl->gfx_manager->deleteAllGraphics();
    pimpl->sound_manager->clear();
}

//////////////////////////////////////////////
// Other methods
//////////////////////////////////////////////

void KnightsGame::setupGfxResizer()
{
    shared_ptr<GfxResizer> gfx_resizer_nearest_nbr(new GfxResizerNearestNbr);
    shared_ptr<GfxResizer> gfx_resizer_scale2x(new GfxResizerScale2x);
    shared_ptr<GfxResizer> gfx_resizer_comp(new GfxResizerCompose(gfx_resizer_nearest_nbr,
                                              pimpl->options->use_scale2x ? gfx_resizer_scale2x : shared_ptr<GfxResizer>(),
                                              CfgInt("lock_to_integer")));
    pimpl->gfx_manager->setGfxResizer(gfx_resizer_comp);
}

void KnightsGame::setupControllers()
{
    pimpl->left_controller.reset(new KeyboardController(pimpl->options->ctrls[0][0],
                                                        pimpl->options->ctrls[0][3],
                                                        pimpl->options->ctrls[0][1],
                                                        pimpl->options->ctrls[0][2],
                                                        pimpl->options->ctrls[0][4],
                                                        pimpl->options->ctrls[0][5],
                                                        pimpl->window));
    pimpl->right_controller.reset(new KeyboardController(pimpl->options->ctrls[1][0],
                                                         pimpl->options->ctrls[1][3],
                                                         pimpl->options->ctrls[1][1],
                                                         pimpl->options->ctrls[1][2],
                                                         pimpl->options->ctrls[1][4],
                                                         pimpl->options->ctrls[1][5],
                                                         pimpl->window));
}    

shared_ptr<Coercri::Font> KnightsGame::getFont() const
{
    return pimpl->font;
}

const Controller & KnightsGame::getLeftController() const
{
    return *pimpl->left_controller;
}

const Controller & KnightsGame::getRightController() const
{
    return *pimpl->right_controller;
}

GfxManager & KnightsGame::getGfxManager() const
{
    return *pimpl->gfx_manager;
}

SoundManager & KnightsGame::getSoundManager() const
{
    return *pimpl->sound_manager;
}

void KnightsGame::getDesktopResolution(int &w, int &h) const
{
    Coercri::GfxDriver::DisplayMode mode = pimpl->gfx_driver->getDesktopMode();
    w = mode.width;
    h = mode.height;
}

const Options & KnightsGame::getOptions() const
{
    return *pimpl->options;
}

void KnightsGame::setAndSaveOptions(const Options &opts)
{
    *pimpl->options = opts;
    if (!pimpl->options_filename.empty()) {
        ofstream str(pimpl->options_filename.c_str(), std::ios_base::out | std::ios_base::trunc);
        SaveOptions(*pimpl->options, str);
    }
    setupControllers();
    setupGfxResizer();
}

//////////////////////////////////////////////
// Main Routine (runKnights)
//////////////////////////////////////////////

void KnightsGame::runKnights()
{
    // Go to title screen
    auto_ptr<Screen> title_screen(new TitleScreen);
    requestScreenChange(title_screen);
    executeScreenChange();

    // Error Handling system.
    string error;
    int num_errors = 0;
    
    // Main Loop
    while (pimpl->running) {

        bool did_something = false;
        unsigned int delay = 0;
        
        if (!error.empty() && num_errors > 100) {
            // Too many errors
            // Abort game.
            throw UnexpectedError("FATAL: " + error);
        }
        
        try {
            if (!error.empty()) {
                // Error detected: try to go to ErrorScreen.
                // (If going to ErrorScreen itself throws an error, then num_errors will keep increasing
                // and eventually we will abort, see above.)
                auto_ptr<Screen> screen;
                screen.reset(new ErrorScreen(error));
                requestScreenChange(screen);
                executeScreenChange();
                error.clear();
                ++num_errors;
                did_something = true;
            }

            // Handle screen changes if necessary
            executeScreenChange();

            // Empty out all event queues
            while (pimpl->gfx_driver->pollEvents()) did_something = true;
            while (pimpl->cg_listener->processInput()) did_something = true;
            while (net_driver && net_driver->doEvents()) did_something = true;
            
            // Do updates
            if (pimpl->update_interval > 0) {
                const unsigned int time_now = pimpl->timer->getMsec();

                // make sure the game doesn't lag too much behind real
                // time, if we don't have enough cpu time to process all
                // updates. (do this by dropping excess updates if
                // necessary.)
                const unsigned int max_lag = pimpl->current_screen->getMaxLag();
                if (time_now - pimpl->last_time > max_lag) {
                    pimpl->last_time = time_now - max_lag;
                }

                // run as many updates as needed to "catch up" to the present
                if (time_now - pimpl->last_time >= pimpl->update_interval) {

                                while (pimpl->gfx_driver->pollEvents()) did_something = true;
            while (pimpl->cg_listener->processInput()) did_something = true;
            while (net_driver && net_driver->doEvents()) did_something = true;

                    
                    pimpl->current_screen->update();
                    pimpl->last_time += pimpl->update_interval;
                    did_something = true;
                }
            
                // Work out how long we have to wait until the next update.
                delay = pimpl->timer->getMsec() - pimpl->last_time;
            }

            // Draw the screen if necessary
            if (pimpl->window->needsRepaint()) {
                std::auto_ptr<Coercri::GfxContext> gc = pimpl->window->createGfxContext();
                pimpl->current_screen->draw(*gc);
                pimpl->cg_listener->draw(*gc);
                pimpl->window->cancelInvalidRegion();
                did_something = true;
            }

            // Sleep if necessary (so that we don't consume 100% CPU).
            if (!did_something) {
                // We never sleep more than 20ms (to maintain GUI responsiveness)
                // nor do we sleep less than 1ms (to ensure that there is always *some* delay).
                if (delay > 20u) delay = 20u;
                if (delay < 1u) delay = 1u;
                pimpl->timer->sleepMsec(delay);
                num_errors = 0;  // if we can get to a sleep w/o error, then assume things
                                 // are going well & can reset num_errors.
            }

        } catch (KConfig::KConfigError& e) {
            // These are big and better displayed on stdout than in a dialog box:
            throw;
        } catch (std::exception &e) {
            error = string("ERROR: ") + e.what();
        } catch (gcn::Exception &e) {
            error = string("Guichan Error: ") + e.getMessage();
        } catch (...) {
            error = "Unknown Error";
        }
    }
        
    // Delete things -- this is to make sure that destructors run in the order we want them to.
    pimpl->current_screen.reset();
    pimpl->requested_screen.reset();
    pimpl->left_controller.reset();
    pimpl->right_controller.reset();
    remote.reset();
    server_config.reset();
    client_config.reset();
    pimpl->sound_driver.reset();
    pimpl->window.reset();
    pimpl->wcl.reset();
    pimpl->cg_listener.reset();
    pimpl->gfx_driver.reset();
}
