/*
 * server_config.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "action.hpp"
#include "anim.hpp"
#include "cfg.hpp"
#include "control.hpp"
#include "create_quest.hpp"
#include "dungeon_generator.hpp"
#include "event_manager.hpp"
#include "gfx_manager.hpp"
#include "gore_manager.hpp"
#include "monster_manager.hpp"
#include "server_config.hpp"
#include "item_generator.hpp"
#include "menu_constraint.hpp"
#include "menu_selections.hpp"
#include "misc/my_exceptions.hpp"
#include "monster_definitions.hpp"
#include "monster_manager.hpp"
#include "quest.hpp"
#include "round.hpp"
#include "rstream/rstream.hpp"
#include "segment.hpp"
#include "special_tiles.hpp"

#include "kconfig/kfile.hpp"
#include "kconfig/kconfig_loader.hpp"
using namespace KConfig;

#include <sstream>
using namespace std;

namespace {
    template<class T> struct Delete {
        void operator()(pair<const Value *const,T*> &x) { delete x.second; }
    };

    void MakeLowerCase(string &s) {
        for (int i=0; i<s.size(); ++i) {
            s[i] = tolower(s[i]);
        }
    }

    // for "MyActionPars" it is expected that the argument list is on the top
    // of the KFile stack.
    class MyActionPars : public ActionPars {
    public:
        MyActionPars(ServerConfig &sc_) : sc(sc_), lst(*sc_.getKFile(), "ActionParameters") { }
        virtual ~MyActionPars() { }
        virtual void require(int n1, int n2) {
            if (lst.getSize() != n1 && (n2<0 || lst.getSize() != n2)) error();
        }
        virtual int getSize() { return lst.getSize(); }
        virtual const Action * getAction(int i) { lst.push(i); return sc.popAction(); }
        virtual int getInt(int i) { lst.push(i); return sc.getKFile()->popInt(); }
        virtual const ItemType *getItemType(int i) { lst.push(i); return sc.popItemType(); }
        virtual int getProbability(int i) { lst.push(i); return sc.popProbability(); }
        virtual const RandomInt * getRandomInt(int i)
            { lst.push(i); return sc.getKFile()->popRandomInt(sc.getRandomIntContainer(), 0); }
        virtual string getString(int i) { lst.push(i); return sc.getKFile()->popString(); }
        virtual shared_ptr<Tile> getTile(int i) { lst.push(i); return sc.popTile(); }
        virtual void error() { sc.getKFile()->error("error while processing action"); }
    private:
        ServerConfig &sc;
        KFile::List lst;
    };

    struct CheckUsage {
        CheckUsage(const vector<bool> &vb_, KFile &kf_, const char *msg1_, const char *msg2_)
            : vb(vb_), kf(kf_), msg1(msg1_), msg2(msg2_) { }
        void operator()(const pair<string,int> &p) { 
            if (vb.at(p.second) == false) {
                kf.error(msg1 + " '" + p.first + "' " + msg2);
            }
        }
        vector<bool> vb;
        KFile &kf;
        string msg1, msg2;
    };

}

ServerConfig::Popper::~Popper()
{
    kf.pop();
}

ServerConfig::ServerConfig(GfxManager &gm)
    : knight_anim(0), default_item(0),
      menu_centre(0), menu_empty(0), menu_highlight(0),
      stuff_bag_graphic(0),
      winner_image(0), loser_image(0),
      blood_icon(0),
      bat_speed(0), zombie_speed(0),
      bat_health(0), zombie_health(0),
      bat_anim(0), zombie_anim(0),
      zombie_weapon(0),
      zombie_ai_hit(0), zombie_ai_fear(0),
      snd_callback(0), snd_cxt(0),
      gfx_manager(gm)
{
}

ServerConfig::~ServerConfig()
{
    for_each(actions.begin(), actions.end(), Delete<Action>());
    for_each(anims.begin(), anims.end(), Delete<Anim>());
    for_each(controls.begin(), controls.end(), Delete<Control>());
    for_each(dungeon_directives.begin(), dungeon_directives.end(), Delete<DungeonDirective>());
    for_each(dungeon_layouts.begin(), dungeon_layouts.end(), Delete<DungeonLayout>());
    for_each(item_generators.begin(), item_generators.end(), Delete<ItemGenerator>());
    for_each(item_types.begin(), item_types.end(), Delete<ItemType>());
    for (size_t i=0; i<special_item_types.size(); ++i) delete special_item_types[i];
    for_each(menu_items.begin(), menu_items.end(), Delete<MenuItem>());
    for_each(menu_ints.begin(), menu_ints.end(), Delete<MenuInt>());
    for_each(menu_values.begin(), menu_values.end(), Delete<MenuValue>());
    for_each(overlays.begin(), overlays.end(), Delete<Overlay>());
    for_each(random_dungeon_layouts.begin(), random_dungeon_layouts.end(),
             Delete<RandomDungeonLayout>());
    for_each(segments.begin(), segments.end(), Delete<Segment>());
    // NB -- No need to do anything for graphics, tiles, random_ints, segment_set or menu.
}

ServerConfig::Sentry::~Sentry()
{
    // Free up memory
    cfg.menu_item_names.clear();
    cfg.segment_categories.clear();
    cfg.tile_categories.clear();
    vector<bool> tmp, tmp2, tmp3, tmp4;
    cfg.segment_categories_used.swap(tmp);
    cfg.segment_categories_defined.swap(tmp2);
    cfg.tile_categories_used.swap(tmp3);
    cfg.tile_categories_defined.swap(tmp4);

    // Close the file
    cfg.kf = shared_ptr<KFile>();
}

namespace {
    // Implementation of KConfigSource to use RStreams...
    struct MyFileLoader : KConfigSource {
        virtual shared_ptr<istream> openFile(const string &filename) {
            shared_ptr<istream> result(new RStream(filename));
            return result;
        }
    };
}

void ServerConfig::loadConfigFile(const string &config_filename,
                                  void (*callback)(const string&,void*), void *cxt)
{
    Sentry s(*this);

    snd_callback = callback;
    snd_cxt = cxt;
    
    // Open the file
    MyFileLoader my_file_loader;
    kf.reset(new KFile(config_filename, my_file_loader, random_ints));

    // Load the Dungeon Environment Customization Menu
    kf->pushSymbol("MAIN_MENU");
    popMenu();
    kf->pushSymbol("MAIN_MENU_SPACE");
    popMenuSpace();
    
    // Load the segment set
    kf->pushSymbol("DUNGEON_SEGMENTS");
    popSegmentSet();

    // Load misc other stuff.
    kf->pushSymbol("WALL");
    wall = popTile();
    kf->pushSymbol("HORIZ_DOOR_1");
    horiz_door[0] = popTile();
    kf->pushSymbol("HORIZ_DOOR_2");
    horiz_door[1] = popTile();
    kf->pushSymbol("VERT_DOOR_1");
    vert_door[0] = popTile();
    kf->pushSymbol("VERT_DOOR_2");
    vert_door[1] = popTile();
    kf->pushSymbol("KNIGHT_ANIM");
    knight_anim = popAnim();
    kf->pushSymbol("KNIGHT_HOUSE_COLOURS");
    popHouseColours();
    kf->pushSymbol("DEFAULT_ITEM");
    default_item = popItemType(0);
    kf->pushSymbol("CONTROLS");
    popControlSet(control_set);
    kf->pushSymbolOptional("CHEAT_CONTROLS");
    if (kf->isNone()) kf->pop(); else popControlSet(cheat_controls);

    // monsters, gore
    kf->pushSymbol("BLOOD_ICON");
    blood_icon = popGraphic();
    kf->pushSymbol("BLOOD_TILES");
    popTileList(blood_tiles);
    kf->pushSymbol("DEAD_KNIGHT_TILES");
    popTileList(dead_knight_tiles);
    kf->pushSymbol("DEAD_BAT_TILES");
    popTileList(dead_bat_tiles);
    kf->pushSymbol("DEAD_ZOMBIE_TILE");
    dead_zombie_tile = popTile();
    kf->pushSymbol("OTHER_ZOMBIE_TILES");
    popTileList(other_zombie_tiles);
    kf->pushSymbol("BAT_PIT_TILES");
    popTileList(bat_pit_tiles);

    // create some monster types...
    kf->pushSymbol("BAT_HEALTH");
    const RandomInt * health = kf->popRandomInt(random_ints, 0);
    kf->pushSymbol("BAT_SPEED");
    int speed = kf->popInt();
    kf->pushSymbol("BAT_ANIM");
    const Anim *anim = popAnim(true);
    kf->pushSymbol("BAT_ATTACK_DAMAGE");
    int ba_dmg = kf->popInt();
    kf->pushSymbol("BAT_ATTACK_STUN_TIME");
    const RandomInt * ba_stun = kf->popRandomInt(random_ints, 0);
    vampire_bat_type.reset(new VampireBatMonsterType(health, speed, anim, ba_dmg, ba_stun));
    kf->pushSymbol("ZOMBIE_HEALTH");
    health = kf->popRandomInt(random_ints, 0);
    kf->pushSymbol("ZOMBIE_SPEED");
    speed = kf->popInt();
    kf->pushSymbol("ZOMBIE_WEAPON");
    const ItemType *weapon = popItemType();
    kf->pushSymbol("ZOMBIE_ANIM");
    anim = popAnim();
    zombie_type.reset(new ZombieMonsterType(health, speed, weapon, anim));
    
    // ai info. Note these could be tweaked/generalized if wanted...
    kf->pushSymbol("ZOMBIE_AI_AVOID");
    if (kf->isNone()) kf->pop(); else popTileList(zombie_ai_avoid);
    kf->pushSymbol("ZOMBIE_AI_HIT");
    zombie_ai_hit = popItemType(0);
    kf->pushSymbol("ZOMBIE_AI_FEAR");
    zombie_ai_fear = popItemType(0);
    
    // misc other stuff
    kf->pushSymbol("SKULL_SETUP");
    popSkullRenderer();
    kf->pushSymbol("POTION_SETUP");
    popPotionRenderer();
    kf->pushSymbol("MENU_CENTRE");
    menu_centre = popGraphic();
    kf->pushSymbol("MENU_EMPTY");
    menu_empty = popGraphic();
    kf->pushSymbol("MENU_HIGHLIGHT");
    menu_highlight = popGraphic();
    kf->pushSymbol("STUFF_BAG_GRAPHIC");
    stuff_bag_graphic = popGraphic();
    kf->pushSymbol("OVERLAY_OFFSETS");
    int oo[N_OVERLAY*3];
    popOverlayOffsets(oo);
    InitOverlayOffsets(oo);

    // winner/loser images
    kf->pushSymbol("WINNER_IMAGE");
    winner_image = popGraphic();
    kf->pushSymbol("LOSER_IMAGE");
    loser_image = popGraphic();

    // Hooks -- These are used for various sound effects that can't
    // be fitted in more directly.
    const char * hook_names[] = {
        "HOOK_WEAPON_DOWNSWING",
        "HOOK_WEAPON_PARRY",
        "HOOK_KNIGHT_DAMAGE",
        "HOOK_CREATURE_SQUELCH",
        "HOOK_BAT",
        "HOOK_ZOMBIE",
        "HOOK_SHOOT",
        "HOOK_MISSILE_MISS"
    };
        
    for (int i=0; i<sizeof(hook_names)/sizeof(hook_names[0]); ++i) {
        const string hook_name(hook_names[i]);
        kf->pushSymbolOptional(hook_name);
        const Action *ac = popAction(0);
        if (ac) {
            hooks.insert(make_pair(hook_name, ac));
        }
    }
    
    // Check that the stack is properly empty
    if (!kf->isStackEmpty()) {
        kf->error("internal error: stack non-empty");
    }
    
    // Check that all segment & tile categories were used
    segment_categories_used.resize(segment_categories.size());  // make sure our bool vectors are big enough
    segment_categories_defined.resize(segment_categories.size());  
    tile_categories_used.resize(tile_categories.size());  
    tile_categories_defined.resize(tile_categories.size());
    for_each(segment_categories.begin(), segment_categories.end(), CheckUsage(segment_categories_used, *kf,
                                                                        "segment category", "is not used by any dungeon directive"));
    for_each(segment_categories.begin(), segment_categories.end(), CheckUsage(segment_categories_defined, *kf,
                                                                              "segment category", "is undefined"));
    for_each(tile_categories.begin(), tile_categories.end(), CheckUsage(tile_categories_used, *kf,
                                                                        "tile category", "is not used by any dungeon directive"));
    for_each(tile_categories.begin(), tile_categories.end(), CheckUsage(tile_categories_defined, *kf,
                                                                        "tile category", "is undefined"));
}


//
// getKnightAnim
//

int ServerConfig::getNumKnightAnims() const
{
    return house_colours_normal.size();
}

const Anim * ServerConfig::getKnightAnim(int i) const
{
    if (!knight_anim) return 0;
    
    if (knight_anims.empty()) {
        // Create the Knight_Anims
        for (int i=0; i<house_colours_normal.size(); ++i) {
            knight_anims.push_back(new Anim(*knight_anim));
            knight_anims.back()->setColourChangeNormal(house_colours_normal[i]);
            knight_anims.back()->setColourChangeInvulnerable(house_colours_invulnerable[i]);
        }
    }
    
    if (i < 0 || i >= getNumKnightAnims()) return 0;
    return knight_anims[i];
}

shared_ptr<const ColourChange> ServerConfig::getSecuredCC(int i) const
{
    return secured_cc.at(i);
}


//
// "Pop" routines
//

MapAccess ServerConfig::popAccessCode(MapAccess dflt)
{
    if (!kf) return A_BLOCKED;

    // We use "Popper" and getString rather than just calling
    // popString directly. This ensures that the string remains on the
    // top of the stack, at the point where "errExpected" is called.
    // This ensures in turn that the correct error traceback is given.

    // Note that this is a slight hack -- a better solution would be
    // to use sentry objects (like KFile::List or KFile::Table) for
    // strings (or other types) as needed. The reason it's a hack is
    // that if (eg) we find an integer, we'll get two errExpected's:
    // one expecting a string and one expecting an access code (in
    // this case).

    Popper p(*kf);
    string s = kf->getString("");

    if (s == "") return dflt;
    else if (s == "approach" || s == "partial") return A_APPROACH;
    else if (s == "blocked") return A_BLOCKED;
    else if (s == "clear") return A_CLEAR;
    kf->errExpected("access code ('approach', 'blocked' or 'clear')");
    return dflt;
}

void ServerConfig::popAccessTable(MapAccess acc[], MapAccess dflt)
{
    if (!kf) return;
    KFile::Table tab(*kf, "AccessTable");
    tab.push("flying");
    acc[H_FLYING] = popAccessCode(dflt);
    tab.push("missiles");
    acc[H_MISSILES] = popAccessCode(dflt);
    tab.push("walking");
    acc[H_WALKING] = popAccessCode(dflt);
}

Action * ServerConfig::popAction()
{
    if (!kf) return 0;
    const Value * p = kf->getTop();
    map<const Value *,Action*>::const_iterator it = actions.find(p);
    if (it == actions.end()) {

        if (kf->isList()) {
            // ListAction

            ListAction *action = new ListAction;
            it = actions.insert(make_pair(p, action)).first;
            
            KFile::List lst(*kf, "Action");
            action->reserve(lst.getSize());
            for (int i=0; i<lst.getSize(); ++i) {
                lst.push(i);
                action->add(popAction(0));
            }

        } else if (kf->isRandom()) {
            // RandomAction

            RandomAction *action = new RandomAction;
            it = actions.insert(make_pair(p, action)).first;

            KFile::Random ran(*kf, "Action");
            action->reserve(ran.getSize());
            for (int i=0; i<ran.getSize(); ++i) {
                int weight = ran.push(i);
                action->add(popAction(), weight);
            }
            
        } else if (kf->isDirective()) {
            // Normal Action

            it = actions.insert(pair<const Value *,Action*>(p, 0)).first;
            
            KFile::Directive dir(*kf, "Action");
            if (!dir.getName().empty()) {
                try {
                    dir.pushArgList();
                    MyActionPars pars(*this);
                    actions.erase(p);
                    Action *ac = ActionMaker::createAction(dir.getName(), pars);
                    if (!ac) {
                        kf->error("unknown action: " + dir.getName());
                        return 0;
                    }
                    it = actions.insert(make_pair(p, ac)).first;

                    // Call back for PlaySound.
                    if (dir.getName() == "PlaySound" && snd_callback) {
                        (*snd_callback)(pars.getString(0), snd_cxt);
                    }
                } catch (BadAction) {
                    kf->errExpected("Action");
                    return 0;
                }
            } else {
                kf->error("directive has no name??");
            }

        } else if (kf->isNone()) {
            // Null Action
            kf->pop();
            return 0;
        } else {
            kf->errExpected("Action");
            kf->pop();
            return 0;
        }
    } else {
        kf->pop();
    }

    if (it->second == 0) {
        kf->error("Recursive action. Try enclosing the action in square brackets?");
    }
    
    return it->second;
}

Action * ServerConfig::popAction(Action *dflt)
{
    if (!kf) return 0;
    if (kf->isNone()) {
        kf->pop();
        return dflt;
    } else {
        return popAction();
    }
}

Anim * ServerConfig::popAnim(bool batmode)
{
    if (!kf) return 0;
    const Value * p = kf->getTop();
    map<const Value *,Anim*>::const_iterator it = anims.find(p);
    if (it == anims.end()) {
        it = anims.insert(make_pair(p, new Anim(batmode))).first;
        KFile::List lst(*kf, "Anim", 4, 12, 16, 32);
        for (int f=0; f<8; ++f) {
            if (lst.getSize() >= (f+1)*4) {
                for (int d=0; d<4; ++d) {
                    lst.push(f*4+d);
                    Coercri::Graphic *g = popGraphic();
                    it->second->setGraphic(MapDirection(d), f, g);
                }
            }
        }
    } else {
        kf->pop();
    }
    return it->second;
}

Anim * ServerConfig::popAnim(Anim *dflt, bool batmode)
{
    if (!kf) return false;
    if (kf->isNone()) {
        kf->pop();
        return dflt;
    } else {
        return popAnim(batmode);
    }
}

BlockType ServerConfig::popBlockType()
{
    if (!kf) return BT_BLOCK;
    
    Popper p(*kf);
    string s = kf->getString("");

    MakeLowerCase(s);
    if (s == "block") return BT_BLOCK;
    else if (s == "edge") return BT_EDGE;
    else if (s == "special") return BT_SPECIAL;
    else if (s == "none") return BT_NONE;
    kf->errExpected("'block', 'edge', 'special' or 'none'");
    return BT_BLOCK;
}

bool ServerConfig::popBool()
{
    if (!kf) return false;

    Popper p(*kf);
    int i = kf->getInt();

    if (i == 0) return false;
    else if (i == 1) return true;
    kf->errExpected("0 or 1");
    return false;
}

bool ServerConfig::popBool(bool dflt)
{
    if (!kf) return false;
    if (kf->isNone()) {
        kf->pop();
        return dflt;
    } else {
        return popBool();
    }
}

Control * ServerConfig::popControl()
{
    if (!kf) return 0;

    const Value * p = kf->getTop();
    map<const Value *,Control*>::iterator it = controls.find(p);
    if (it == controls.end()) {
        it = controls.insert(make_pair(p, new Control)).first;
        KFile::Table tab(*kf, "Control");   
        tab.push("action");
        Action *ac = popAction();
        tab.push("continuous");
        bool cts = popBool(false);
        tab.push("menu_direction");
        MapDirection mdir = popMapDirection(D_NORTH);
        tab.push("menu_icon");
        Coercri::Graphic *g = popGraphic(0);
        tab.push("menu_special");
        unsigned int ms = kf->popInt(0);
        tab.push("tap_priority");
        int tp = kf->popInt(0);

        it->second->construct(ac, g, mdir, tp, cts, ms);
    }

    return it->second;
}

Control * ServerConfig::popControl(Control *dflt)
{
    if (!kf) return 0;
    if (kf->isNone()) {
        kf->pop();
        return dflt;
    } else {
        return popControl();
    }
}

void ServerConfig::popControlSet(vector<const Control*> &which_control_set)
{
    if (!kf) return;
    KFile::List lst(*kf, "ControlSet");
    which_control_set.reserve(lst.getSize());
    for (int i=0; i<lst.getSize(); ++i) {
        lst.push(i);
        which_control_set.push_back(popControl());
    }
}

ServerConfig::DungeonBlock ServerConfig::popDungeonBlock()
{
    if (!kf) return DungeonBlock();
    DungeonBlock result;
    KFile::Table tbl(*kf, "DungeonBlock");
    tbl.push("exits");
    result.exits = kf->popString("nesw");
    tbl.push("type");
    result.bt = popBlockType();
    return result;
}

DungeonLayout * ServerConfig::doPopDungeonLayout(int w, int h)
{
    if (!kf) return 0;
    KFile::List lst(*kf, "DungeonLayoutData", w*h);  // final arg gives expected number of list entries.

    DungeonLayout *dlay = new DungeonLayout(w, h);
    try {
        vector<DungeonBlock> db(w*h);
        for (int x=0; x<w; ++x) {
            for (int y=0; y<h; ++y) {
                lst.push(y*w+x);
                db[y*w+x] = popDungeonBlock();
                dlay->setBlockType(x, y, db[y*w+x].bt);
            }
        }
        
        for (int x=0; x<w; ++x) {
            for (int y=0; y<h-1; ++y) {
                if (db[y*w+x].bt == BT_NONE || db[(y+1)*w+x].bt == BT_NONE) continue;
                bool dn = db[y    *w+x].exits.find_first_of("sS") != string::npos;
                bool up = db[(y+1)*w+x].exits.find_first_of("nN") != string::npos;
                if (up != dn) kf->error("exits do not match");
                dlay->setVertExit(x, y, dn);
            }
        }
        
        for (int x=0; x<w-1; ++x) {
            for (int y=0; y<h; ++y) {
                if (db[y*w+x].bt == BT_NONE || db[y*w+(x+1)].bt == BT_NONE) continue;
                bool rt = db[y*w+x    ].exits.find_first_of("eE") != string::npos;
                bool lf = db[y*w+(x+1)].exits.find_first_of("wW") != string::npos;
                if (lf != rt) kf->error("exits do not match");
                dlay->setHorizExit(x, y, lf);
            }
        }

        return dlay;
        
    } catch (...) {
        delete dlay;
        throw;
    }

}

DungeonDirective * ServerConfig::popDungeonDirective()
{
    if (!kf) return 0;
    
    const Value * p = kf->getTop();
    map<const Value *,DungeonDirective*>::iterator it = dungeon_directives.find(p);
    if (it == dungeon_directives.end()) {
        KFile::Directive dir(*kf, "MenuDirective");
        dir.pushArgList();
        DungeonDirective *d = DungeonDirective::create(dir.getName(), *this); // pops arglist
        it = dungeon_directives.insert(make_pair(p, d)).first;
    }
    return it->second;
}


DungeonLayout * ServerConfig::popDungeonLayout()
{
    if (!kf) return 0;
    const Value * p = kf->getTop();
    map<const Value *, DungeonLayout*>::iterator it = dungeon_layouts.find(p);
    if (it == dungeon_layouts.end()) {
        KFile::Table tab(*kf, "DungeonLayout");
        tab.push("height",false);
        int h = kf->popInt();
        tab.push("width",false);
        int w = kf->popInt();

        tab.reset();        
        tab.push("data");
        DungeonLayout *dlay = doPopDungeonLayout(w, h);
        tab.push("height"); kf->pop();
        tab.push("width"); kf->pop();

        it = dungeon_layouts.insert(make_pair(p, dlay)).first;
    }

    return it->second;
}


Coercri::Graphic * ServerConfig::popGraphic()
{
    if (!kf) return 0;
    const Value * p = kf->getTop();
    map<const Value *,Coercri::Graphic*>::iterator it = graphics.find(p);
    if (it == graphics.end()) {
        KFile::List lst(*kf, "Graphic", 1, 4, 6);

        lst.push(0);
        string filename = kf->popString();

        int x = 0, y = 0, r = -1, g = -1, b = -1;
        
        if (lst.getSize() > 1) {
            lst.push(1);
            r = kf->popInt();
            lst.push(2);
            g = kf->popInt();
            lst.push(3);
            b = kf->popInt();
            if (lst.getSize() > 4) {
                lst.push(4);
                x = kf->popInt();
                lst.push(5);
                y = kf->popInt();
            }
        }

        Coercri::Graphic *gfx = 0;
        try {
            gfx = gfx_manager.createGraphicFromFile(filename, x, y, r, g, b);
        } catch (GraphicLoadFailed &glf) {
            kf->error(glf.what());
        }
        it = graphics.insert(make_pair(p, gfx)).first;
    }

    return it->second;
}

Coercri::Graphic * ServerConfig::popGraphic(Coercri::Graphic *dflt)
{
    if (!kf) return 0;
    if (kf->isNone()) {
        kf->pop();
        return dflt;
    } else {
        return popGraphic();
    }
}

void ServerConfig::popHouseColours()
{
    // Here we set the local variable "house_colours" (both normal and invulnerable versions).
    // Expected input format is a list of lists of RGB colours.
    if (!kf) return;
    KFile::List lst(*kf, "HouseColourList");
    if (lst.getSize() < 1) {
        kf->error("house colour list is empty");
    } else {

        Coercri::Color rgb;  // work space
        vector<Coercri::Color> src_cols;  // colours of the first House
        
        // First entry: N "src" colors (store these in src_cols)
        {
            lst.push(0);
            KFile::List cols(*kf, "HouseColours");
            for (int i=0; i<cols.getSize(); ++i) {
                cols.push(i);
                rgb = popRGB();
                src_cols.push_back(rgb);
            }
            // first House's ColourChange is always empty
            house_colours_normal.push_back(ColourChange());
            // for first House's invulnerability, we set a single colour change
            // from src_cols[0] to CfgInt("invuln_r/g/b")
            ColourChange cc;
            if (!src_cols.empty()) {
                cc.add(src_cols[0], Coercri::Color(CfgInt("invuln_r"), CfgInt("invuln_g"), CfgInt("invuln_b")));
            }
            house_colours_invulnerable.push_back(cc);
        }
        
        // Second and subsequent entries: N "dest" colors
        // (create a ColourChange from "src_cols" to the newly-read dest cols)
        // for the Invulnerable one: we overwrite the first entry with the value of
        // CfgInt("invuln_r/g/b").
        for (int n=1; n<lst.getSize(); ++n) {
            lst.push(n);
            KFile::List cols(*kf, "HouseColours", src_cols.size());
            ColourChange cc, cc2;
            for (int i=0; i<cols.getSize(); ++i) {
                cols.push(i);
                rgb = popRGB();
                cc.add(src_cols[i], rgb);
                cc2.add(src_cols[i], i==0? Coercri::Color(CfgInt("invuln_r"), CfgInt("invuln_g"), CfgInt("invuln_b")) : rgb);
            }
            house_colours_normal.push_back(cc);
            house_colours_invulnerable.push_back(cc2);
        }

        // Now do the secured cc's:
        // (By convention, home CC's change from red ie $FF0000 to the first colour
        // in the house colours list)
        for (int i=0; i<lst.getSize(); ++i) {
            lst.push(i);
            KFile::List cols(*kf, "HouseColours");
            Coercri::Color rgb;
            cols.push(0);
            rgb = popRGB();
            shared_ptr<ColourChange> cc(new ColourChange);
            cc->add(Coercri::Color(255,0,0),rgb);
            secured_cc.push_back(cc);
        }
    }
}

ItemGenerator * ServerConfig::popItemGenerator()
{
    if (!kf) return 0;

    const Value * p = kf->getTop();
    map<const Value *, ItemGenerator*>::iterator itor = item_generators.find(p);

    if (itor == item_generators.end()) {
        ItemGenerator *ig = new ItemGenerator;
        itor = item_generators.insert(make_pair(p, ig)).first;
        if (kf->isTable()) {
            ig->setFixedItemType(popItemType(), 0);
        } else if (kf->isList()) {
            KFile::List lst(*kf, "ItemAndQuantity", 2);
            lst.push(0);
            const ItemType *it = popItemType();
            lst.push(1);
            const RandomInt * qty = kf->popRandomInt(random_ints, 0);
            ig->setFixedItemType(it, qty);
        } else {
            KFile::Random ran(*kf, "ItemGenerator");
            ig->reserve(ran.getSize());
            for (int i=0; i<ran.getSize(); ++i) {
                int wt = ran.push(i);
                ItemGenerator *subig = popItemGenerator();
                ig->add(subig, wt);
            }
        }

    } else {
        kf->pop();
        if (itor->second == 0) {
            kf->error("error: possible recursion in ItemGenerator");
        }
    }

    return itor->second;
}

ItemSize ServerConfig::popItemSize()
{
    if (!kf) return IS_NOPICKUP;

    Popper p(*kf);
    string s = kf->getString("");

    if (s == "held") return IS_BIG;
    else if (s == "backpack") return IS_SMALL;
    else if (s == "magic") return IS_MAGIC;
    else if (s == "nopickup") return IS_NOPICKUP;
    kf->errExpected("'backpack', 'held', 'magic' or 'nopickup'");
    return IS_NOPICKUP;
}

ItemSize ServerConfig::popItemSize(ItemSize dflt)
{
    if (!kf) return IS_NOPICKUP;
    if (kf->isNone()) {
        kf->pop();
        return dflt;
    } else {
        return popItemSize();
    }
}

ItemType * ServerConfig::popItemType()
{
    if (!kf) return 0;

    const Value * p = kf->getTop();
    map<const Value *, ItemType*>::iterator itor = item_types.find(p);
    if (itor == item_types.end()) {
        KFile::Table tab(*kf, "ItemType");
    
        ItemType *it = new ItemType;
        itor = item_types.insert(make_pair(p, it)).first;

        tab.push("allow_strength");
        bool allow_strength = popBool(true);
        
        tab.push("ammo");
        const ItemType *ammo = popItemType(0);
        
        tab.push("backpack_graphic");
        Coercri::Graphic *backpack_graphic = popGraphic(0);

        tab.push("backpack_overdraw");
        Coercri::Graphic *backpack_overdraw = popGraphic(0);
        
        tab.push("backpack_slot");
        int backpack_slot = kf->popInt(0);

        tab.push("can_throw");
        bool can_throw = popBool(false);

        tab.push("control");
        Control *control = popControl(0);

        tab.push("editor_label"); kf->pop();
        
        tab.push("fragile");
        bool fragile = popBool(false);

        tab.push("graphic");
        Coercri::Graphic *graphic = popGraphic(0);
        
        tab.push("key");
        int key = kf->popInt(0);
        
        tab.push("max_stack");
        int max_stack = kf->popInt(1);


        // Melee properties
        tab.push("melee_action");
        const Action *melee_action = popAction(0);
        
        bool any = false, all = true;
        tab.push("melee_backswing_time");
        any |= (!kf->isNone()); all &= (!kf->isNone());
        int melee_backswing_time = kf->popInt(0);

        tab.push("melee_damage");
        any |= (!kf->isNone()); all &= (!kf->isNone());
        const RandomInt * melee_damage = kf->popRandomInt(random_ints, 0);
        
        tab.push("melee_downswing_time");
        any |= (!kf->isNone()); all &= (!kf->isNone());     
        int melee_downswing_time = kf->popInt(0);
        
        tab.push("melee_stun_time");
        any |= (!kf->isNone()); all &= (!kf->isNone());     
        const RandomInt * melee_stun_time = kf->popRandomInt(random_ints, 0);
        
        tab.push("melee_tile_damage");
        any |= (!kf->isNone()); all &= (!kf->isNone());     
        const RandomInt * melee_tile_damage = kf->popRandomInt(random_ints, 0);

        if (any && !all) {
            kf->error("not all melee properties were given");
        }


        // Missile properties
        any = false; all = true;
        if (can_throw) any = true;
        
        tab.push("missile_access_chance");
        any |= (!kf->isNone()); all &= (!kf->isNone());     
        int missile_acc_ch = popProbability(0);
        
        tab.push("missile_anim");
        any |= (!kf->isNone()); all &= (!kf->isNone());     
        const Anim * missile_anim = popAnim((Anim*)0);
        
        tab.push("missile_backswing_time");
        if (can_throw) { any |= (!kf->isNone()); all &= (!kf->isNone()); }
        int missile_backswing_time = kf->popInt(0);
        
        tab.push("missile_damage");
        any |= (!kf->isNone()); all &= (!kf->isNone());
        const RandomInt *missile_damage = kf->popRandomInt(random_ints, 0);
        
        tab.push("missile_downswing_time");
        if (can_throw) { any |= (!kf->isNone()); all &= (!kf->isNone()); }
        int missile_downswing_time = kf->popInt(0);
        
        tab.push("missile_hit_multiplier");
        any |= (!kf->isNone()); all &= (!kf->isNone());     
        int missile_hit_multiplier = kf->popInt(1);
        
        tab.push("missile_range");
        any |= (!kf->isNone()); all &= (!kf->isNone());     
        int missile_range = kf->popInt(0);
        
        tab.push("missile_speed");
        any |= (!kf->isNone()); all &= (!kf->isNone());     
        int missile_speed = kf->popInt(0);
        
        tab.push("missile_stun_time");
        any |= (!kf->isNone()); all &= (!kf->isNone());     
        const RandomInt * missile_stun_time = kf->popRandomInt(random_ints, 0);

        if (any && !all) {
            kf->error("not all missile properties were given");
        }
        
        
        tab.push("on_drop");
        Action *on_drop = popAction(0);
        tab.push("on_hit");
        Action *on_hit = popAction(0);
        tab.push("on_pick_up");
        Action *on_pick_up = popAction(0);
        tab.push("on_walk_over");
        Action *on_walk_over = popAction(0);
        
        tab.push("open_traps");
        bool open_traps = popBool(false);

        tab.push("overlay");
        Overlay *overlay = popOverlay(0);
        
        tab.push("parry_chance");
        int parry_chance = popProbability(0);

        tab.push("reload_action");
        const Action * reload_action = popAction(0);

        tab.push("reload_action_time");
        int reload_action_time = kf->popInt(250);
        
        tab.push("reload_time");
        int reload_time = kf->popInt(0);
        
        tab.push("stack_graphic");
        Coercri::Graphic *stack_graphic = popGraphic(graphic);
        
        tab.push("type");
        ItemSize item_size = popItemSize(IS_NOPICKUP);
        
        it->construct(graphic, stack_graphic, backpack_graphic, backpack_overdraw, overlay, item_size, max_stack,
                      backpack_slot, fragile,
                      melee_backswing_time, melee_downswing_time, melee_damage,
                      melee_stun_time, melee_tile_damage, melee_action, parry_chance, 
                      can_throw, missile_range, missile_speed, missile_acc_ch,
                      missile_hit_multiplier, missile_backswing_time, missile_downswing_time,
                      missile_damage, missile_stun_time, missile_anim,
                      reload_time, ammo, reload_action, reload_action_time,
                      key, open_traps, control, on_pick_up, on_drop, on_walk_over, on_hit,
                      allow_strength);

        // If a crossbow, then also set up a second itemtype for the loaded version.
        if (it->canLoad()) {
            ItemType *it2 = new ItemType(*it);   // make a copy of the old itemtype
            special_item_types.push_back(it2);
            it->setLoaded(it2);
            it2->setUnloaded(it);
        }

    } else {
        kf->pop();
    }
    
    return itor->second;
}

ItemType * ServerConfig::popItemType(ItemType *dflt)
{
    if (!kf) return 0;
    if (kf->isNone()) {
        kf->pop();
        return dflt;
    } else {
        return popItemType();
    }
}

MapDirection ServerConfig::popMapDirection()
{
    if (!kf) return D_NORTH;

    Popper p(*kf);
    string s = kf->getString("");

    MakeLowerCase(s);
    if (s == "north" || s == "up") return D_NORTH;
    else if (s == "east" || s == "right") return D_EAST;
    else if (s == "south" || s == "down") return D_SOUTH;
    else if (s == "west" || s == "left") return D_WEST;
    kf->errExpected("compass direction");
    return D_NORTH;
}

MapDirection ServerConfig::popMapDirection(MapDirection dflt)
{
    if (!kf) return D_NORTH;
    if (kf->isNone()) {
        kf->pop();
        return dflt;
    } else {
        return popMapDirection();
    }
}

void ServerConfig::popMenu()
{
    if (!kf) return;

    KFile::List lst(*kf, "Menu");

    for (int i=0; i<lst.getSize(); ++i) {
        lst.push(i);
        if (kf->isDirective()) {
            DungeonDirective *d = popDungeonDirective();
            menu.addDungeonDirective(d);
        } else if (kf->isTable()) {
            MenuItem * mi = popMenuItem();
            menu.addItem(mi);
        } else {
            kf->errExpected("MenuItem");
            kf->pop();
        }
    }
}

void ServerConfig::popMenuSpace()
{
    if (!kf) return;

    KFile::List lst(*kf, "MenuSpace");

    for (int i=0; i<lst.getSize(); ++i) {
        lst.push(i);
        const string id = kf->popString();
        menu.addSpaceAfter(id);
    }
}

MenuItem * ServerConfig::popMenuItem()
{
    if (!kf) return 0;
    
    const Value * p = kf->getTop();
    map<const Value *, MenuItem*>::iterator it = menu_items.find(p);

    if (it == menu_items.end()) {
        KFile::Table tab(*kf, "MenuItem");
    
        tab.push("id");
        string id = kf->popString();
        for (map<const Value *,MenuItem*>::const_iterator it2 = menu_items.begin();
        it2 != menu_items.end(); ++it2) {
            if (it2->second->getID() == id) {
                kf->error(string("duplicate menu item: ") + id);
                break;
            }
        }
        it = menu_items.insert(make_pair(p, new MenuItem)).first;
        it->second->setID(id);

        tab.push("min_value");
        it->second->setMinValue( kf->popInt(0) );

        tab.push("title");
        it->second->setTitle( kf->popString() );

        tab.push("values");
        KFile::List lst(*kf, "MenuValues");
        for (int i=0; i<lst.getSize(); ++i) {
            lst.push(i);
            MenuValue *val = popMenuValue();
            it->second->addValue(val);
        }
    }
    return it->second;
}

MenuInt * ServerConfig::popMenuInt()
{
    if (!kf) return 0;
    const Value * p = kf->getTop();
    map<const Value *, MenuInt*>::iterator it = menu_ints.find(p);

    if (it == menu_ints.end()) {
        it = menu_ints.insert(make_pair(p, static_cast<MenuInt*>(0))).first;
        if (kf->isInt()) {
            int val = kf->popInt();
            it->second = new MenuIntConst(val);
        } else if (kf->isDirective()) {
            KFile::Directive dir(*kf, "GetMenuDirective");
            if (dir.getName() != "GetMenu") {
                kf->errExpected("integer or GetMenu directive");
            } else {
                dir.pushArgList();
                KFile::List lst(*kf, "string", 1);
                if (lst.getSize()==1) {
                    lst.push(0);
                    string menu_item_name = kf->popString();
                    it->second = new MenuIntVar(menu_item_name);
                }
            }
        } else {
            kf->errExpected("integer or GetMenu directive");
            kf->pop();
        }

    } else {
        kf->pop();
    }

    return it->second;
}

MenuValue * ServerConfig::popMenuValue()
{
    if (!kf) return 0;
    const Value * p = kf->getTop();
    map<const Value *, MenuValue*>::iterator it = menu_values.find(p);
    
    if (it == menu_values.end()) {
        it = menu_values.insert(make_pair(p, new MenuValue)).first;

        if (kf->isList()) {
            KFile::List lst(*kf, "MenuValue");
            lst.push(0);
            popMenuValueName(*it->second);
            for (int i=1; i<lst.getSize(); ++i) {
                lst.push(i);
                popMenuValueDirective(*it->second);
            }
        } else {
            popMenuValueName(*it->second);
        }
    } else {
        kf->pop();
    }
    return it->second;
}

void ServerConfig::popMenuValueDirective(MenuValue &mv)
{
    if (!kf) return;
    KFile::Directive dir(*kf, "MenuDirective");
    string dname = dir.getName();
    if (dname.substr(0,7) == "Dungeon") {
        DungeonDirective *d = popDungeonDirective();
        if (d) mv.addDungeonDirective(d);
    } else if (dname.substr(0,5) == "Quest") {
        dir.pushArgList();
        shared_ptr<Quest> q = CreateQuest(dname, *this);
        if (q) mv.addQuest(q);
    } else if (dname.substr(0,9) == "Constrain") {
        dir.pushArgList();
        shared_ptr<MenuConstraint> c = CreateMenuConstraint(dname, *this);
        if (c) mv.addConstraint(c);
    } else {
        kf->errExpected("MenuDirective");
    }
}


void ServerConfig::popMenuValueName(MenuValue &mv)
{
    if (!kf) return;
    if (kf->isString()) {
        mv.setText( kf->popString() );
    } else if (kf->isDirective()) {
        KFile::Directive dir(*kf, "MenuValueName");
        if (dir.getName() == "MenuNumber") {
            // do nothing
        } else if (dir.getName() != "MenuIcon") {
            kf->errExpected("MenuValueName");
        } else {
            dir.pushArgList();
            KFile::List lst(*kf, "string or graphic", 1);
            if (lst.getSize()==1) {
                lst.push(0);
                if (kf->isString()) {
                    mv.setCharIcon( kf->popString()[0] );
                } else {
                    mv.setGraphicIcon( popGraphic() );
                }
            }
        }
    } else {
        kf->errExpected("MenuValueName");
        kf->pop();
    }
}


Overlay* ServerConfig::popOverlay()
{
    if (!kf) return 0;
    
    const Value * p = kf->getTop();
    map<const Value *, Overlay*>::iterator it = overlays.find(p);

    if (it == overlays.end()) {
        KFile::List lst(*kf, "Overlay", 4);

        if (lst.getSize()==4) {
            it = overlays.insert(make_pair(p, new Overlay)).first;
            for (int i=0; i<4; ++i) {
                lst.push(i);
                it->second->setGraphic(MapDirection(i), popGraphic(0));
            }
        }
    } else {
        kf->pop();
    }
    return it->second;
}

Overlay * ServerConfig::popOverlay(Overlay *dflt)
{
    if (!kf) return 0;
    if (kf->isNone()) {
        kf->pop();
        return dflt;
    } else {
        return popOverlay();
    }
}

void ServerConfig::popOverlayOffsets(int *p)
{
    if (!kf) return;
    KFile::List lst(*kf, "OverlayOffsetList", N_OVERLAY*3);
    for (int i=0; i<N_OVERLAY; ++i) {
        lst.push(i*3);
        *p++ = popMapDirection();
        lst.push(i*3+1);
        *p++ = kf->popInt();
        lst.push(i*3+2);
        *p++ = kf->popInt();
    }
}
        
void ServerConfig::popPotionRenderer()
{
    if (!kf) return;
    KFile::Table tab(*kf, "PotionSetup");
    {
        tab.push("colours");
        KFile::List lst(*kf, "PotionColours");
        for (int i=0; i<lst.getSize(); ++i) {
            lst.push(i);
            potion_renderer.addColour(popRGB());
        }
    }   
    {
        tab.push("graphics");
        KFile::List lst(*kf, "PotionGraphics");
        for (int i=0; i<lst.getSize(); ++i) {
            lst.push(i);
            potion_renderer.addGraphic(popGraphic());
        }
    }
}

int ServerConfig::popProbability()
{
    if (!kf) return 0;

    Popper pp(*kf);
    int p = kf->getInt();

    if (p < 0 || p > 100) {
        kf->errExpected("percentage");
        p = 0;
    }
    return p;
}

int ServerConfig::popProbability(int dflt)
{
    if (!kf) return 0;
    if (kf->isNone()) {
        kf->pop();
        return dflt;
    } else {
        return popProbability();
    }
}

RandomDungeonLayout * ServerConfig::popRandomDungeonLayout()
{
    if (!kf) return 0;

    const Value * p = kf->getTop();
    map<const Value *, RandomDungeonLayout*>::iterator it = random_dungeon_layouts.find(p);
    if (it == random_dungeon_layouts.end()) {
        RandomDungeonLayout *r = new RandomDungeonLayout;
        it = random_dungeon_layouts.insert(make_pair(p, r)).first;
        if (kf->isTable()) {
            DungeonLayout *d = popDungeonLayout();
            r->add(d);
        } else if (kf->isRandom()) {
            KFile::Random ran(*kf, "DungeonLayout");
            for (int i=0; i<ran.getSize(); ++i) {
                int weight = ran.push(i);
                DungeonLayout *d = popDungeonLayout();
                for (int j=0; j<weight; ++j) r->add(d);
            }
        }
        
    } else {
        kf->pop();
    }

    return it->second;
}   

Coercri::Color ServerConfig::popRGB()
{
    if (!kf) return Coercri::Color(0,0,0,0);
    int x = kf->popInt();
    return Coercri::Color((x >> 16) & 255,  (x >> 8) & 255, x & 255);
}

Segment * ServerConfig::popSegment(const vector<SegmentTileData> &tile_map, int &category)
{
    if (!kf) return 0;
    KFile::Table tab(*kf, "Segment");

    tab.push("height",false);
    int h = kf->popInt();
    tab.push("width",false);
    int w = kf->popInt();
    tab.reset();

    tab.push("bat_placement_tile");
    int t = kf->popInt(0);
    shared_ptr<Tile> bat_tile;
    if (t != 0) {
        // Bats will be placed on the *first* tile in the tile-list
        // for the given number.
        const SegmentTileData & st = tileFromInt(tile_map, t);
        if (!st.tiles.empty()) bat_tile = st.tiles.front();
    }
    
    tab.push("category");
    string catname = kf->popString("");
    category = getSegmentCategory(catname);
    defineSegmentCategory(category);  // this segment category exists.
    tab.push("data");
    Segment *segment = popSegmentData(tile_map, w, h);
    if (!segment) return 0;
    segment->setBatPlacementTile(bat_tile);
    tab.push("height"); kf->pop();
    tab.push("name"); kf->pop();  // used only by the map editor
    tab.push("rooms");
    popSegmentRooms(*segment);
    tab.push("switches"); 
    if (kf->isNone()) kf->pop();
    else popSegmentSwitches(*segment);
    tab.push("width"); kf->pop();
    return segment;
} 

const ServerConfig::SegmentTileData & ServerConfig::tileFromInt(const vector<SegmentTileData> &tile_map, int t)
{
    if (t < 1 || t > tile_map.size()) {
        static SegmentTileData err;
        kf->error("tile number out of range!");
        return err;
    } else {
        return tile_map[t-1];
    }
}
    
void ServerConfig::processTile(const vector<SegmentTileData> &tile_map, Segment &segment,
                               int x, int y)
{
    const int tile = kf->popInt();
    const SegmentTileData & seg_tile(tileFromInt(tile_map, tile));
    for (int i=0; i<seg_tile.tiles.size(); ++i) {
        segment.addTile(x, y, seg_tile.tiles[i]);
    }
    for (int i=0; i<seg_tile.items.size(); ++i) {
        segment.addItem(x, y, seg_tile.items[i]);
    }
}

Segment * ServerConfig::popSegmentData(const vector<SegmentTileData> &tile_map,
                                       int w, int h)
{
    if (!kf) return 0;
    if (w<=0 || h<=0) {
        kf->error("bad width/height");
        kf->pop();
        return 0;
    } else {
        KFile::List lst(*kf, "SegmentData", w*h);
        Segment *segment = new Segment(w, h);
        try {
            for (int y=0; y<h; ++y) {
                for (int x=0; x<w; ++x) {
                    lst.push(y*w+x);
                    processTile(tile_map, *segment, x, y);  // pops
                }
            }
            return segment;
        } catch (...) {
            delete segment;
            throw;
        }
    }
}

void ServerConfig::popSegmentList(const vector<SegmentTileData> &tile_map)
{
    if (!kf) return;
    
    KFile::List lst(*kf, "SegmentList");

    for (int i=0; i<lst.getSize(); ++i) {
        lst.push(i);
        int rcat;
        Segment *rm = popSegment(tile_map, rcat);
        if (rm) segment_set.addSegment(rm, rm->getHomes().size(), rcat);
    }
}

void ServerConfig::popSegmentRooms(Segment &s)
{
    if (!kf) return;
    KFile::List lst(*kf, "RoomList");
    for (int i=0; i<lst.getSize(); ++i) {
        lst.push(i);
        KFile::List rum(*kf, "RoomData", 4);
        rum.push(0);
        const int tlx = kf->popInt()-1;
        rum.push(1);
        const int tly = kf->popInt()-1;
        rum.push(2);
        const int w = kf->popInt();
        rum.push(3);
        const int h = kf->popInt();
        if (tlx < -1 || tly < -1 || w <= 0 || h <= 0 || tlx+w-1 > s.getWidth() || tly+h-1 > s.getHeight()) kf->error("bad room data");
        s.addRoom(tlx, tly, w, h);
    }
}

void ServerConfig::popSegmentSet()
{
    if (!kf) return;

    KFile::Table tbl(*kf,"SegmentSet");

    vector<SegmentTileData> tile_map;
    tbl.push("tiles",false);
    popSegmentTiles(tile_map);

    tbl.reset();
    tbl.push("segments");
    popSegmentList(tile_map);
    tbl.push("tiles");
    kf->pop();
}

void ServerConfig::popSegmentSwitches(Segment &r)
{
    if (!kf) return;
    KFile::List lst(*kf, "SwitchList");
    for (int i=0; i<lst.getSize(); ++i) {
        lst.push(i);
        KFile::List sw(*kf, "SwitchData", 4);
        sw.push(0);
        bool activate = popBool();
        sw.push(1);
        int x = kf->popInt();
        sw.push(2);
        int y = kf->popInt();
        sw.push(3);
        Action *ac = popAction();
        shared_ptr<Tile> t(new Tile); // dummy tile for the switch-action.
        t->construct(activate?0:ac, activate?ac:0);
        r.addTile(x,y,t);
    }
}

void ServerConfig::popSegmentTiles(vector<SegmentTileData> &tile_map)
{
    if (!kf) return;
    KFile::List lst(*kf, "TileTable");
    
    for (int i=0; i<lst.getSize(); ++i) {
        SegmentTileData seg_tile;
        bool items_yet = false;
        lst.push(i);

        if (kf->isList()) {
            KFile::List lst2(*kf, "SegmentTile");
            for (int j=0; j<lst2.getSize(); ++j) {
                lst2.push(j);
                popSegmentTile(seg_tile, items_yet);
            }
        } else {
            popSegmentTile(seg_tile, items_yet);
        }
        tile_map.push_back(seg_tile);
    }
}

void ServerConfig::popSegmentTile(SegmentTileData &seg_tile, bool &items_yet)
{
    if (items_yet) {
        const ItemType *itype = popItemType();
        seg_tile.items.push_back(itype);
    } else {
        if (kf->isInt()) {
            // This is a slight hack: You put an integer (usually '0') to signify
            // the end of tiles and the beginning of items. 
            items_yet = true;
            kf->pop();
        } else {
            shared_ptr<Tile> t = popTile();
            if (t) seg_tile.tiles.push_back(t);
        }
    }
}
    
void ServerConfig::popSkullRenderer()
{
    if (!kf) return;
    KFile::Table tab(*kf, "SkullSetup");
    {
        tab.push("columns");
        KFile::List lst(*kf, "SkullColumns");
        for (int i=0; i<lst.getSize(); ++i) {
            lst.push(i);
            skull_renderer.addColumn(kf->popInt());
        }
    }
    {
        tab.push("graphics");
        KFile::List lst(*kf, "SkullGraphics");
        for (int i=0; i<lst.getSize(); ++i) {
            lst.push(i);
            skull_renderer.addGraphic(popGraphic());
        }
    }
    {
        tab.push("rows");
        KFile::List lst(*kf, "SkullRows");
        for (int i=0; i<lst.getSize(); ++i) {
            lst.push(i);
            skull_renderer.addRow(kf->popInt());
        }
    }
}

ServerConfig::StairInfo ServerConfig::popStairsDown()
{
    StairInfo result;
    result.is_stair = result.special = false;
    result.down_direction = D_NORTH;
    if (kf) {
        if (kf->isNone()) {
            kf->pop();
        } else if (kf->isString() && kf->getString()=="special") {
            kf->pop();
            result.special = true;
        } else {
            result.is_stair = true;
            result.down_direction = popMapDirection();
        }
    }
    return result;
}

shared_ptr<Tile> ServerConfig::popTile()
{
    if (!kf) return shared_ptr<Tile>();

    const Value * p = kf->getTop();
    map<const Value *, shared_ptr<Tile> >::iterator it = tiles.find(p);

    if (it == tiles.end()) {
        KFile::Table tab(*kf, "Tile");

        shared_ptr<Tile> tile;
        shared_ptr<Home> home;
        shared_ptr<Door> door;
        shared_ptr<Chest> chest;
        tab.push("type", false);
        string ttype = kf->popString("");
        tab.reset();
        if (ttype == "") {
            tile.reset(new Tile);
        } else if (ttype == "home") {
            home.reset(new Home);
            tile = home;
        } else if (ttype == "door") {
            door.reset(new Door);
            tile = door;
        } else if (ttype == "chest") {
            chest.reset(new Chest);
            tile = chest;
        } else if (ttype == "barrel") {
            tile.reset(new Barrel);
        } else if (ttype == "no_special_open") {
            tile.reset(new NoSpecialOpenTile);
        } else {
            kf->error("unknown tile type: " + ttype);
        }

        it = tiles.insert(make_pair(p, tile)).first;
        if (!tile) return tile; // unk tile type -- don't try to go any further.

        MapAccess acc[H_MISSILES+1];
        tab.push("access");
        popAccessTable(acc, A_BLOCKED);

        tab.push("can_trap");
        const bool can_trap = popBool(true);
        if (door && !can_trap) door->setCannotTrap();
        if (chest && !can_trap) chest->setCannotTrap();
        
        tab.push("control");
        Control *ctrl = popControl(0); 

        tab.push("depth");
        int depth = kf->popInt(0);

        tab.push("editor_label"); kf->pop();

        MapDirection facing = D_NORTH;
        if (chest || home) {
            tab.push("facing");
            facing = popMapDirection(facing);
        }
        
        tab.push("graphic");
        Coercri::Graphic *graphic = popGraphic(0);

        tab.push("hit_points");
        const RandomInt *initial_hit_points = kf->popRandomInt(random_ints, 0);
        
        tab.push("items");
        int item_category = popTileItems(acc);
        bool items_allowed = (item_category > -2); // -2 means "no items allowed", -3 means "destroy items"
        bool destroy_items = (item_category == -3);

        int lock_chance = 0, lock_pick_only_chance = 0, keymax = 0;
        if (door || chest) {
            tab.push("keymax");
            keymax = kf->popInt(1);
            tab.push("lock_chance");
            lock_chance = kf->popInt(0);
            tab.push("lock_pick_only_chance");
            lock_pick_only_chance = kf->popInt(0);
        }

        tab.push("on_activate");
        const Action *on_activate = popAction(0);
        tab.push("on_approach");
        const Action *on_approach = popAction(0);
        tab.push("on_close");
        const Action *on_close = popAction(0);
        tab.push("on_destroy");
        const Action *on_destroy = popAction(0);
        tab.push("on_hit");
        const Action *on_hit = popAction(0);
        tab.push("on_open");
        const Action *on_open = popAction(0);
        tab.push("on_walk_over");
        const Action *on_walk_over = popAction(0);
        tab.push("on_withdraw");
        const Action *on_withdraw = popAction(0);
        
        bool open = false;
        Coercri::Graphic *open_graphic = 0;
        if (door) {
            // Chests are always closed initially. (This is assumed here, as well as
            // in the Chest class probably.)
            tab.push("open");
            open = popBool(false);
        }
        if (door || chest) {
            tab.push("open_graphic");
            open_graphic = popGraphic(0);
        }

        // doors/chests can have "special_lock" property
        bool special_locked = false;
        if (door || chest) {
            tab.push("special_lock");
            special_locked = popBool(false);
        }
            
        tab.push("stairs_down");
        StairInfo stair_info = popStairsDown();

        // chests can have traps
        int trap_chance = 0;
        vector<Chest::TrapInfo> traps;
        if (chest) {
            tab.push("trap_chance");
            trap_chance = popProbability();
            tab.push("traps");
            KFile::List lst(*kf, "TrapsList");
            for (int i=0; i<lst.getSize(); ++i) {
                lst.push(i);
                KFile::List lst2(*kf, "TrapDefList", 2);
                lst2.push(0);
                const ItemType *itype = popItemType();
                lst2.push(1);
                const Action *ac = popAction();
                traps.push_back(Chest::TrapInfo(itype, ac));
            }
        }
        
        tab.push("type"); kf->pop();

        Coercri::Color uns_col;
        if (home) {
            tab.push("unsecured_colour");
            uns_col = popRGB();
        }
                
        // For open doors, we need to override the access codes and set them all to A_CLEAR
        // (backup orig. access codes first, for doorConstruct)
        MapAccess acc2[H_MISSILES+1];
        for (int i=0; i<=H_MISSILES; ++i) acc2[i] = acc[i];
        if (open) {
            for (int i=0; i<=H_MISSILES; ++i) {
                acc[i] = A_CLEAR;
            }
        }

        tile->construct(open ? open_graphic : graphic, depth,
                        open ? true : items_allowed, open ? false : destroy_items,
                        item_category, acc,
                        stair_info.is_stair, stair_info.special,
                        stair_info.down_direction,
                        initial_hit_points, on_destroy, on_activate, on_close, on_open,
                        on_walk_over, on_approach, on_withdraw, on_hit, ctrl);
        if (open) {
            ASSERT(door);
            door->setOpenInitially();
        }
        
        if (chest) {
            chest->chestConstruct(open_graphic, graphic, facing, trap_chance, traps);
            chest->setLockChance(lock_chance, lock_pick_only_chance, keymax);
            if (special_locked) chest->setSpecialLock();
        } else if (door) {
            door->doorConstruct(open_graphic, graphic, acc2);
            door->setLockChance(lock_chance, lock_pick_only_chance, keymax);
            if (special_locked) door->setSpecialLock();
        } else if (home) {
            shared_ptr<ColourChange> cc(new ColourChange);
            cc->add(Coercri::Color(255,0,0), uns_col);
            home->homeConstruct(facing, cc);
        }
        
    } else {
        kf->pop();
    }

    return it->second;
}

shared_ptr<Tile> ServerConfig::popTile(shared_ptr<Tile> dflt)
{
    if (!kf) return shared_ptr<Tile>();
    if (kf->isNone()) {
        kf->pop();
        return dflt;
    } else {
        return popTile();
    }
}

// returns:
// -3 if items are to be destroyed (pits)
// -2 if items not allowed here
// -1 if items allowed, but none are to be generated
// 0 or higher: items to be generated, the number gives the index of the tile-category to use.
int ServerConfig::popTileItems(const MapAccess acc[])
{
    if (!kf) return -2;
    if (kf->isInt()) {
        bool b = popBool();
        return b ? -1 : -2;
    } else if (kf->isNone()) {
        // default (allow items for A_CLEAR, else block)
        kf->pop();
        if (acc[H_WALKING] == A_CLEAR) return -1;
        else return -2;
    } else if (kf->isString()) {
        // string, naming a tile category
        // return the number corresponding to that category.
        string s = kf->popString();
        if (s == "destroy") {
            return -3;
        } else {
            int tc = getTileCategory(s);
            defineTileCategory(tc);  // this category is used in some tile (not just referenced by someone).
            return tc;
        }
    } else {
        kf->errExpected("'0', '1', item-category or 'destroy'");
        kf->pop();
        return 0;
    }
}

void ServerConfig::popTileList(vector<shared_ptr<Tile> > &output)
{
    output.clear();
    if (!kf) return;
    KConfig::KFile::List lst(*kf, "TileList");
    output.reserve(lst.getSize());
    for (int i=0; i<lst.getSize(); ++i) {
        lst.push(i);
        output.push_back(popTile());
    }
}


int ServerConfig::popTime()
{
    if (!kf) return 0;

    Popper p(*kf);
    float t = kf->getFloat();

    if (t < 0.0f) {
        kf->errExpected("non-negative number");
        t = 0.0f;
    }
    return Round(t * CfgInt("ticks_per_second"));
}

int ServerConfig::popTime(int dflt)
{
    if (!kf) return 0;
    if (kf->isNone()) {
        kf->pop();
        return dflt;
    } else {
        return popTime();
    }
}

int ServerConfig::getSegmentCategory(const string &cname) 
{
    if (cname.empty()) return -1;
    return segment_categories.insert(make_pair(cname, segment_categories.size())).first->second;
}

int ServerConfig::getTileCategory(const string &cname)
{
    if (cname.empty()) return -1;
    return tile_categories.insert(make_pair(cname, tile_categories.size())).first->second;
}

void ServerConfig::doUse(int item, int sz, vector<bool> &container)
{
    if (item < 0 || item >= sz) return;
    if (container.size() < sz) {
        container.resize(sz);
    }
    container[item] = true;
}

void ServerConfig::serializeAll(Coercri::NetworkConnection &remote)
{
    // overlay offsets
    SerializeOverlayOffsets(remote);
    
    // graphics
    remote.putLong(graphics.size());
    int next_id = 1;
    for (map<const Value *, Coercri::Graphic*>::iterator it = graphics.begin(); it != graphics.end();
    ++it) {
        gfx_manager.serializeGraphic(remote, *it->second, next_id++);
    }

    // anims (includes knight anims)
    getKnightAnim(0); // make sure we're initialized
    remote.putLong(anims.size() + knight_anims.size());
    next_id = 1;
    for (map<const Value *, Anim*>::iterator it = anims.begin(); it != anims.end(); ++it) {
        it->second->serialize(remote, gfx_manager, next_id++);
    }
    for (int i=0; i<knight_anims.size(); ++i) {
        knight_anims[i]->serialize(remote, gfx_manager, next_id++);
    }

    // overlays
    remote.putLong(overlays.size());
    next_id = 1;
    for (map<const Value *, Overlay*>::iterator it = overlays.begin(); it != overlays.end();
    ++it) {
        it->second->serialize(remote, gfx_manager, next_id++);
    }

    // controls
    for (int i=0; i<NUM_STANDARD_CONTROLS; ++i) {
        controls_by_id.push_back(g_standard_controls[i]);
    }
    for (map<const Value *, Control*>::iterator it = controls.begin(); it != controls.end();
    ++it) {
        controls_by_id.push_back(it->second);
    }
    remote.putLong(NUM_STANDARD_CONTROLS);
    remote.putLong(controls_by_id.size() - NUM_STANDARD_CONTROLS);
    for (int i=NUM_STANDARD_CONTROLS; i<controls_by_id.size(); ++i) {
        controls_by_id[i]->serialize(remote, gfx_manager, i+1);
    }

    // potion, skull renderers
    potion_renderer.serialize(remote, gfx_manager);
    skull_renderer.serialize(remote, gfx_manager);

    // menu gfx
    remote.putLong(menu_centre? gfx_manager.getGraphicID(*menu_centre) : 0);
    remote.putLong(menu_empty? gfx_manager.getGraphicID(*menu_empty) : 0);
    remote.putLong(menu_highlight? gfx_manager.getGraphicID(*menu_highlight) : 0);

    // winner/loser images
    remote.putLong(winner_image? gfx_manager.getGraphicID(*winner_image) : 0);
    remote.putLong(loser_image? gfx_manager.getGraphicID(*loser_image) : 0);
}

const Control * ServerConfig::getControlFromID(int n) const
{
    if (n < 1 || n > controls_by_id.size()) return 0;
    else return controls_by_id[n-1];
}

void ServerConfig::configureGoreManager(GoreManager &gm, const vector<Player*> &players) const
{
    if (players.size() < 2) return;
    if (dead_knight_tiles.size() < 4) return;
    gm.setKnightCorpse(*players[0], dead_knight_tiles[0], dead_knight_tiles[1]);
    gm.setKnightCorpse(*players[1], dead_knight_tiles[2], dead_knight_tiles[3]);

    for (int i=0; i<dead_bat_tiles.size(); ++i) {
        gm.addMonsterCorpse(vampire_bat_type.get(), dead_bat_tiles[i]);
    }
    gm.addMonsterCorpse(zombie_type.get(), dead_zombie_tile);

    for (int i=0; i<blood_tiles.size(); ++i) {
        gm.addBloodTile(blood_tiles[i]);
    }

    gm.setBloodIcon(blood_icon);
}

void ServerConfig::configureMonsterManager(MonsterManager &mm, int vampire_bats,
                                           int zombie_activity) const
{
    // Vampire bats
    for (int i=0; i<bat_pit_tiles.size(); ++i) {
        mm.addMonsterGenerator(bat_pit_tiles[i], vampire_bat_type.get());
    }
    
    // Zombies
    // -- Dead knights and "other zombie tiles" decay into "dead zombie tile"
    // -- "Dead zombie tile" can be reanimated into a zombie.
    for (int i=0; i<dead_knight_tiles.size(); ++i) {
        mm.addZombieDecay(dead_knight_tiles[i], dead_zombie_tile);
    }
    for (int i=0; i<other_zombie_tiles.size(); ++i) {
        mm.addZombieDecay(other_zombie_tiles[i], dead_zombie_tile);
    }
    mm.addZombieReanimate(dead_zombie_tile, zombie_type.get());

    // Set the zombie and bat activity levels
    mm.setZombieChance(4*zombie_activity*zombie_activity);   // quadratic from 0 to 100%
    mm.setBatChance(20*vampire_bats);  // linear from 0 to 100%
    
    // Monster limits
    // -- Number of bats is limited to three times the vampire bats level
    // -- Number of zombies is not limited
    // -- Total number of monsters is limited to CfgInt("monster_limit").
    mm.limitMonster(vampire_bat_type.get(), vampire_bats*3);
    mm.limitTotalMonsters(CfgInt("monster_limit"));

    // AI
    mm.setZombieAI(zombie_ai_avoid, zombie_ai_hit, zombie_ai_fear);
}

void ServerConfig::configureEventManager(EventManager &evt_mgr) const
{
    evt_mgr.setupHooks(hooks);
}
